package net.usa.rharder.tabusearch22;

/**
 * There is a great deal of flexibility
 * in defining these moves. A move could be toggling
 * a binary variable on or off. It could be swapping
 * two variables in a sequence. The important thing
 * is that it affects the solution in some way and is
 * also able to undo itself immediately afterward.
 * For each move generated by the <code>TSMoveManager</code>
 * the engine <code>operate</code>s on the current solution,
 * <code>evaluate</code>s it with the objective function and
 * penalty function, and then performs <code>undoOperation</code>
 * 
 * @author Robert Harder, rharder@usa.net
 * @see TSFunction
 * @see TSSolution
 * @version 2.0
 * @since 1.0
 */
        
public interface TSMove
{   

    /**
     * The required method <code>operateOn</code> accepts
     * a solution to modify and does so. Note that a new
     * solution is not returned. The original solution
     * is modified.
     * 
     * @param soln The solution to be modified.
     * @see TSSolution
     * @since 1.0
     */
    public abstract void operateOn( TSSolution soln );

    /**
     * A move needs to be able to undo itself. When the
     * <code>TSEngine</code> evaluates a solution, it does not
     * create a copy before a move operates on it. This results
     * in a significant performance increase, but the <code>TSMove</code>
     * needs to be more sophisticated and be able to undo itself.
     * 
     * @param soln Solution to be un-operated on.
     * @see TSSolution
     * @since 1.1
     */
    
    public abstract void undoOperation( TSSolution soln );

    /**
     * If a number of moves do
     * not affect each other, the <code>TSEngine</code> will execute
     * all of these moves at that iteration.
     * <P>
     * For example, in a Traveling Salesman Problem moving customers
     * around in one vehicle's tour probably does not affect
     * moving customers in another vehicle's tour. If this is true
     * for your problem, you can specify that a particular move
     * does or does not conflict with the passed move object.
     * <P>
     * To be considered "not conflicting" the incremental value
     * of making one move must be independent of the incremental
     * value of making another. In other words if <var>delta-X</var>
     * is the extra value (objective function + penalty function)
     * for making move <var>X</var> and <var>delta-Y</var> is the
     * extra value for making move <var>Y</var>, then the new solution's
     * total value must be equal to the old value + <var>delta-X</var>
     * + <var>delta-Y</var>.
     * <P>
     * For traditional tabu search behavior, always return
     * <code>true</code in this method.
     * 
     * @param move The move to compare this one to.
     * @return whether or not the two moves can both be executed at the
     * same iteration
     * @see TSSolution
     * @see TSFunction
     * @since 2.0
     */
    
    public abstract boolean conflictsWith( TSMove move );
                   
}   // end class TSMove
    